from typing import Iterable

from genv.entities import Snapshot
from genv.metrics import Collection as Base, Spec

from ..utils import Host


class Collection(Base):
    """
    A hostname aware metric collection.
    """

    def __init__(self, specs: Iterable[Spec]) -> None:
        return super().__init__(
            [
                Spec(
                    spec.type,
                    spec.name,
                    spec.documentation,
                    ("hostname",) + spec.labelnames,
                    spec.convert,
                    spec.filter,
                )
                for spec in specs
            ]
        )

    def cleanup(self, hosts: Iterable[Host], snapshots: Iterable[Snapshot]) -> None:
        """
        Cleans up metric label values.
        """
        hostname_to_snapshot = {
            host.hostname: snapshot for host, snapshot in zip(hosts, snapshots)
        }

        for metric in self:

            def filter(label_set: Iterable[str]) -> bool:
                hostname, *label_set = label_set

                if hostname not in hostname_to_snapshot:
                    return False

                if metric.filter and not metric.filter(
                    label_set, hostname_to_snapshot[hostname]
                ):
                    return False

                return True

            metric.cleanup(filter)

    def update(self, hosts: Iterable[Host], snapshots: Iterable[Snapshot]) -> None:
        """
        Updates metrics of hosts according to the given snapshots.
        """
        for host, snapshot in zip(hosts, snapshots):
            super().update(snapshot, labels={"hostname": host.hostname})



from contextlib import ExitStack, contextmanager
import subprocess
from typing import Any, Iterable, Union

import genv.utils
from genv.entities import Device, Devices
import genv.serialization

import genv.core.envs

from .utils import File


class State(File[Devices]):
    def __init__(self, cleanup: bool = True, reset: bool = False) -> None:
        super().__init__(
            genv.utils.get_temp_file_path("devices.json"),
            cleanup,
            reset,
        )

    def _create(self):
        # move to genv.utils.nvidia_smi once async is supported here
        devices_total_memory = [
            f"{int(line)}mi"
            for line in subprocess.check_output(
                "GENV_BYPASS=1 nvidia-smi --query-gpu=memory.total --format=csv,noheader,nounits",
                shell=True,
            ).splitlines()
        ]

        return Devices(
            [
                Device(index, total_memory, [])
                for index, total_memory in enumerate(devices_total_memory)
            ]
        )

    def _convert(self, o: Union[Any, Devices]) -> Devices:
        # the following logic converts state files from versions <= 0.8.0.
        # note that the indicator is 'o.devices' and not 'o' itself because the structure of
        # the state file from these versions was similar to the snapshot structure (a single
        # key named "devices") so the JSON decoder parsed the object as a snapshot object.
        if isinstance(o.devices, dict):
            o = Devices(
                [
                    Device(
                        int(index),
                        device["total_memory"],
                        [
                            Device.Attachement(
                                env["eid"],
                                env.get("gpu_memory", None),
                                env["attached"],
                            )
                            for env in device["eids"].values()
                        ],
                    )
                    for index, device in o.devices.items()
                ]
            )

        return o

    def _clean(self, devices: Devices) -> None:
        envs = genv.core.envs.snapshot()

        devices.cleanup(poll_eid=lambda eid: eid in envs.eids)


def snapshot() -> Devices:
    """
    Returns an environments snapshot.
    """
    return State().load()


def attach(eid: str, allow_over_subscription: bool = False) -> Iterable[int]:
    """
    Attaches an environment to devices.
    The device count is taken from the environment configuration.
    Does not detach devices if already attached to more devices.

    :return: Attached device indices
    """
    with State() as devices:
        envs = genv.core.envs.snapshot()
        env_config = envs[eid].config

        if env_config.gpus is not None:
            env_devices = devices.filter(eid=eid)

            diff = env_config.gpus - len(env_devices)

            if diff > 0:
                not_env_devices = devices.filter(not_indices=env_devices.indices)

                indices = not_env_devices.find_available_devices(
                    diff, env_config.gpu_memory, allow_over_subscription
                )

                devices.attach(eid, indices, env_config.gpu_memory)
            elif diff < 0:
                pass  # TODO(raz): support detaching devices if already attached to more

        return devices.filter(eid=eid).indices


def detach(eid: str, index: int) -> None:
    """
    Detaches an environment from a device.
    """
    with State() as devices:
        devices[index].detach(eid)


def get_lock_path(index: int, create: bool = False) -> str:
    """
    Returns the path of a device lock file.
    Creates the file if requested and it does not exist.
    """

    path = genv.utils.get_temp_file_path(f"devices/{index}.lock")

    if create:
        genv.utils.create_lock(path)

    return path


@contextmanager
def lock(indices: Union[Iterable[int], int]) -> None:
    """
    Obtain exclusive access to a device or a few devices.
    """
    if isinstance(indices, int):
        indices = [indices]

    # we sort here to decrease the chance of deadlock when different processes
    # try to lock the same devices in a different order.
    indices = sorted(indices)

    # NOTE(raz): currently, we wait on a lock even if it is already taken by our environment.
    # we should think if this is the desired behavior and if it's possible to lock once per environment.

    with ExitStack() as es:
        for index in indices:
            es.enter_context(genv.utils.access_lock(get_lock_path(index)))

        yield

    # TODO(raz): currently we wait for the entire device to become available.
    # we should support fractional usage and allow multiple environments to
    # access the device if the sum of their memory capacity fits.



import os
from typing import Optional


def eid() -> Optional[str]:
    """
    Returns the current environment identifier or None if not running in one.
    """
    return os.environ.get("GENV_ENVIRONMENT_ID")


def active() -> bool:
    """
    Returns whether running in an active environment.
    """
    return eid() is not None



from dataclasses import dataclass
from typing import Callable, Iterable, Optional, Tuple

from genv.entities import Snapshot

from .type import Type

@dataclass
class Spec:
    """
    A metric specification.
    """

    type: Type
    name: str
    documentation: str
    labelnames: Tuple[str] = ()
    convert: Optional[Callable[[Snapshot], float]] = None
    filter: Optional[Callable[[Iterable[str], Snapshot], bool]] = None



import argparse
import subprocess
import sys
from typing import Iterable, NoReturn

import genv


def do_lock(args: Iterable[str]) -> NoReturn:
    """
    Runs the provided command under the device lock.
    """
    with genv.sdk.devices.lock():
        status = subprocess.run(args).returncode if len(args) else 0
        sys.exit(status)


def parse_args() -> argparse.Namespace:
    """
    Parses the passed arguments.
    """

    def lock(parser: argparse.ArgumentParser):
        parser.add_argument(
            "args",
            nargs=argparse.REMAINDER,
            metavar="COMMAND",
            help="Command to run under the device lock",
        )

    parser = argparse.ArgumentParser(
        description="Query and control Genv on this machine or in a cluster"
    )

    subparsers = parser.add_subparsers(dest="command", metavar="SUBCOMMAND")

    for command, help in [
        (lock, "Lock over-subscribed devices"),
    ]:
        command(subparsers.add_parser(command.__name__, help=help))

    return parser.parse_args()


def main():
    """
    The genvctl entrypoint.
    """

    args = parse_args()

    if args.command == "lock":
        do_lock(args.args)


if __name__ == "__main__":
    main()



from contextlib import contextmanager
from glob import glob
import os
import re
from typing import Iterable, Optional

import genv.utils
import genv.core.devices

import genv.sdk.env


def visible() -> Optional[Iterable[int]]:
    """
    Returns the indices of visible devices if set.
    """

    indices = os.environ.get("CUDA_VISIBLE_DEVICES")

    if indices is None:
        return None

    if indices == "-1":
        return []

    return [int(index) for index in indices.split(",")]


def lockable() -> Iterable[int]:
    """
    Returns the indices of devices that have an existing lock.
    """

    dir = genv.utils.get_temp_file_path("devices")

    return [
        int(re.sub(r".*?(\d+)\.lock", r"\1", path)) for path in glob(f"{dir}/*.lock")
    ]


@contextmanager
def lock() -> None:
    """
    Obtains exclusive access to the attached devices.
    Does nothing if not running in an active environment or not attached to devices.
    """

    if not genv.sdk.env.active():
        yield
    else:
        if "GENV_SHELL" in os.environ:
            indices = visible()
        elif "GENV_CONTAINER" in os.environ:
            indices = lockable()
        else:
            raise RuntimeError("Failed to find device indices to lock")

        with genv.core.devices.lock(indices):
            yield



import asyncio
import os
from asyncio.subprocess import Process
from typing import Dict, Optional, List, Tuple

from .runner import Runner as Base


class Runner(Base):
    host_name: str
    timeout: Optional[int]
    __SSH_COMMAND_PREFIX = "ssh"
    __SSH_TIMEOUT_PARAMETER = "-o ConnectTimeout={0}"

    def __init__(self, host_name: str, timeout: Optional[int] = None, process_env: Optional[Dict] = None):
        super().__init__(process_env)
        self.host_name = host_name
        self.timeout = timeout

    def name(self) -> str:
        return self.host_name

    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        ssh_parameters = self.calc_ssh_params()
        remote_command = self.calc_command_on_remote_machine(args, sudo)

        return await asyncio.create_subprocess_exec(
            Runner.__SSH_COMMAND_PREFIX,
            *ssh_parameters,
            remote_command,
            stdin=stdin_fd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

    def calc_ssh_params(self) -> List[str]:
        ssh_parameters = []
        if self.timeout is not None:
            ssh_parameters.append(self.__SSH_TIMEOUT_PARAMETER.format(self.timeout))
        ssh_parameters.append(self.host_name)
        return ssh_parameters

    def calc_command_on_remote_machine(self, args: Tuple[str, ...], sudo: bool) -> str:
        command = " ".join(args)
        if self._process_env:
            command = self._add_environment_vars(command, self._process_env)
        if sudo:
            command = f"sudo {command}"
        return command

    def _get_error_msg(self, command: str, stderr: str):
        return f"Failed to run a command using ssh on {self.host_name}: command: '{command}' ({stderr})"

    @staticmethod
    def _add_environment_vars(command: str, process_env: Dict[str, str]):
        env_str = " ".join([f'{var_key}={var_value}' for var_key, var_value in process_env.items()])
        command = f'env {env_str} {command}'
        return command

    @staticmethod
    def calc_remote_path_env(root: str) -> str:
        path_env = f"$PATH:{root}/bin"

        if os.path.realpath(os.path.join(os.environ["GENV_ROOT"], "devel/shims")) in [
            os.path.realpath(path) for path in os.environ["PATH"].split(":")
        ]:
            path_env = f"{path_env}:{root}/devel/shims"

        return path_env



from typing import Iterable

from genv.entities import Snapshot

from .metric import Metric
from .spec import Spec
from .type import Type


class Collection:
    """
    A metric collection.
    """

    def __init__(self, specs: Iterable[Spec]) -> None:
        self._metrics = [
            Metric(
                spec.name,
                spec.documentation,
                spec.labelnames,
                type=spec.type,
                convert=spec.convert,
                filter=spec.filter,
            )
            for spec in specs
        ]

    def __iter__(self):
        return self._metrics.__iter__()

    def __getitem__(self, name: str) -> Metric:
        return next(metric for metric in self if metric.name == name)

    def _find(self, type: Type) -> Iterable[Metric]:
        """
        Returns all metrics of the given type.
        """
        return [metric for metric in self if metric.type == type]

    def cleanup(self, snapshot: Snapshot) -> None:
        """
        Cleans up metric label values.
        """
        for metric in self:
            if not metric.filter:
                continue

            metric.cleanup(lambda label_set: metric.filter(label_set, snapshot))

    def update(self, snapshot: Snapshot, labels: dict = {}) -> None:
        """
        Updates metrics according to the given snapshot.
        """
        for group in [
            self._system,
            self._env,
            self._process,
            self._user,
        ]:
            group(snapshot, labels)

    def _system(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates system-wide metrics.
        """
        for metric in self._find(Type.System):
            metric.labels(**labels).update(snapshot)

    def _env(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-environment metrics.
        """
        for env in snapshot.envs:
            env_snapshot = snapshot.filter(eid=env.eid)

            for metric in self._find(Type.Environment):
                metric.labels(eid=env.eid, **labels).update(env_snapshot)

    def _process(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-process metrics.
        """
        for process in snapshot.processes:
            pid = process.pid
            eid = process.eid or ""

            self["genv_process_devices_total"].labels(pid=pid, eid=eid, **labels).set(
                len(process.indices)
            )

            for usage in process.used_gpu_memory:
                self["genv_process_used_gpu_memory_bytes"].labels(
                    pid=pid, eid=eid, device=usage.index, **labels
                ).set(usage.bytes)

    def _user(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-user metrics.
        """
        for username in snapshot.envs.usernames:
            user_snapshot = snapshot.filter(username=username)

            for metric in self._find(Type.User):
                metric.labels(username=username, **labels).update(user_snapshot)



import asyncio
from asyncio.subprocess import Process

from .runner import Runner as Base


class Runner(Base):
    def name(self) -> str:
        return "local"

    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        if sudo:
            args = ["sudo", *args]

        return await asyncio.create_subprocess_exec(
            *args,
            env=self._process_env,
            stdin=stdin_fd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

    def _get_error_msg(self, command: str, stderr: str):
        return f"Failed to run a command on the local machine: command: '{command}' ({stderr})"



import setuptools

with open("../README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="genv",
    version="0.11.0",
    author="Run:AI",
    author_email="pypi@run.ai",
    description="GPU Environment Management",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/run-ai/genv",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": [
            "genvctl = genv.__main__:main",
        ]
    },
)



from typing import Callable, Iterable, Optional

import prometheus_client

from genv.entities import Snapshot

from .spec import Type


class Metric(prometheus_client.Gauge):
    """
    A Genv metric.
    """

    def __init__(
        self,
        name: str,
        documentation: str,
        labelnames: Iterable[str],
        *,
        type: Type,
        convert: Optional[Callable[[Snapshot], float]],
        filter: Optional[Callable[[Iterable[str], Snapshot], bool]],
        **kwargs,
    ):
        super().__init__(name, documentation, labelnames, **kwargs)
        self._kwargs["type"] = self.type = type
        self._kwargs["convert"] = self.convert = convert
        self._kwargs["filter"] = self.filter = filter

    @property
    def name(self) -> str:
        return self._name

    def labels(self, *labelvalues, **labelkwargs):
        """
        Sets metric label values if passed.
        """
        if labelvalues or labelkwargs:
            return super().labels(*labelvalues, **labelkwargs)

        return self

    def cleanup(self, filter: Callable[[Iterable[str]], bool]) -> None:
        """
        Cleans up metric label values.
        """
        with self._lock:
            label_sets = list(self._metrics.keys())

        for label_set in label_sets:
            if not filter(label_set):
                self.remove(*label_set)

    def update(self, snapshot: Snapshot) -> None:
        """
        Updates the metric value according to the given snapshot.
        Note that some metrics expect the snapshot to be already filtered.
        """
        if self.convert is None:
            raise RuntimeError(
                "Conversion function must be provided when using snapshots"
            )

        self.set(self.convert(snapshot))



import asyncio
from abc import ABC, abstractmethod
from asyncio.subprocess import Process
from typing import Dict, Optional


class CommandResults:
    command_process: Process
    stdout: str
    stderr: str

    def __init__(self, command_process: Process, stdout: str, stderr: str):
        self.command_process = command_process
        self.stdout = stdout
        self.stderr = stderr


class Runner(ABC):
    DEFAULT_STD_ENCODING = 'utf-8'
    _process_env: Dict[str, str]

    def __init__(self, process_env: Optional[Dict] = None):
        self._process_env = process_env

    async def run(self, *args: str, stdin: Optional[str] = None, sudo: bool = False, check: bool = False
                  ) -> CommandResults:
        stdin_fd = asyncio.subprocess.PIPE if stdin else asyncio.subprocess.DEVNULL
        process = await self._open_process(*args,
                                           stdin_fd=stdin_fd,
                                           sudo=sudo)

        stdout, stderr = await process.communicate(stdin.encode(Runner.DEFAULT_STD_ENCODING) if stdin else None)
        stdout = stdout.decode(self.DEFAULT_STD_ENCODING).strip()
        stderr = stderr.decode(self.DEFAULT_STD_ENCODING).strip()

        if check and process.returncode != 0:
            command = " ".join(args)
            stdin_str = ' with stdin ' + str(stdin)
            raise RuntimeError(
                f"Failed running '{command}' {' with sudo ' if sudo else ''} { stdin_str if stdin else '' } ({stderr})"
            )

        return CommandResults(process, stdout, stderr)

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError('This should be implemented in subclasses')

    @abstractmethod
    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        raise NotImplementedError('This should be implemented in subclasses')

    @abstractmethod
    def _get_error_msg(self, command: str, stderr: str):
        raise NotImplementedError('This should be implemented in subclasses')
